(* ::Package:: *)

(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Model file: sub-threshold annihilation scenario

constant annihilation and scattering amplitude

Model parameters:

    mDM     -   DM mass
    gDM     -   # of internal degrees of freedom of DM
    g       -   coupling
    r       -   mDMprime/mDM

Version: 12-10-2020
Contact: Tobias Binder (Kavli IPMU)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)


(*************** specification of the model ***************)

Print["----------------- Model: Sub-threshold annihilation scenario ----------------"];

cardinfo:={"-------- card: "<>parameters<>" mDM="<>ToString[N[mDM]]<>"  gDM="<>ToString[gDM]<>" g="<>ToString[g]<>" r="<>ToString[r]<>"----"};


(*--------------------------------annihilation cross section -------------------------*)

sv[s_]                          :=  (g^2/(32 Pi gDM^2)) (Sqrt[s-4 r^2 mDM^2]/(Sqrt[s](s-2 mDM^2)));
resonance                       :=   {};
threshold                       :=   If[r>1,{4*r^2 mDM^2},{}];

(* uncomment for including off-shell decay effects. Specify width.
lambda[y_,z_]:=1+y^2+z^2-2y-2y z -2z;
sv[s_]  := If[s<1.05r^2*4mDM^2,((g^2/(16Pi))/(s-2mDM^2))NIntegrate[Re[Sqrt[lambda[r^2mDM^2/s,mu r^2mDM^2/s]] width*mu/((mu-1)^2+width^2)],{mu,0,(1-Sqrt[s]/(r mDM))^2},Method -> {Automatic, "SymbolicProcessing" -> 0}]/Pi,(g^2/(16 Pi gDM^2)) (Sqrt[s-4 r^2 mDM^2]/(Sqrt[s](s-2 mDM^2)))];
resonance={};
threshold={};
*)

(*-------------------------------- Momentum transfer rate FP approximation  -------------------------*)

gam[x_]                         :=  (mDM g^2/(6 Pi^3)) Exp[- r x] (r^2/((1+r)^4)) x^(-2);
scatttype                       =   "gamma(x)";


modelparameters                 :=  {mDM,g^2,r}; (* modelparameters is passed to the below routines (param) *)


(* Optional routines: *)

GetAnnMatrixAnalytic=Compile[{{pref,_Real,0},{dim,_Integer,0},{c2,_Real,0},{xi,_Real,0},{tq,_Real,1},{param,_Real,1}},Module[{i,j,tr,ctu,matrix,sp,sm,qi,qj,xqi,xqj,q,xq,inter,const,mdm,M2,r},
(*
INPUT:

    pref                        =   [gDM/(2 Pi^2 tilde H x)]*(m/Sqrt[x])^3 dqA for momentum coordinates A,
    pref                        =   [gDM/(2 Pi^2 tilde H x)]*(m/x)^3 g_s dqB for momentum coordinates B.
    dim                         -   Dimesion of tq (momentum table)
    p/T                         =   Sqrt[c2] tq, where c2 and tq depend on momentum coordinates A and B.
    xi                          =   mDM/T
    param                       =   modelparameters

OUTPUT:

    pref*<svmol>_theta(p,tilde p) -- pref*angular averaged annihilation cross section in form of a matrix

COMMENT:

Used in fBE if Analyticsvtheta  = True;
This routine is specialized to the threshold example.
Please modify "model dependent parts" for adopting it to other cases where the angular avarage can be performed analytically.
*)


matrix                          =   ConstantArray[0.,{dim,dim}];
q                               =   Sqrt[c2] tq; (*=p/T*)
xq                              =   Sqrt[xi^2+q^2]; (*=E/T*)


(* Model dependent*)
(*###############*)
mdm                             =   param[[1]];
M2                              =   param[[2]];
r                               =   param[[3]];
tr                              =   4. r^2 xi^2;
const                           =   (xi^2/mdm^2)(1./2.)(1./2.)(1./2.)(M2/(32. 3.1415 ));(*factors 1/2: costheta average, (P1*P2)sigmavmol, costheta to s*)
(*###############*)

Do[
    qi                          =   q[[i]];
    xqi                         =   xq[[i]];

    (* Model dependent*)
    (*###############*)
    ctu                         =   (xqi^2 +xi^2-tr/2.)/(qi^2);
    If[ctu>-1,
        If[ctu>1,
            sp                  =   4. xi^2,
            sp                  =   tr
        ];
            sm                  =   2. xi^2+2.(xqi^2 +qi^2);
        If[sp-tr>=0&&sm-tr>=0&&sm>sp,
            inter               =   (const/(xqi^2 qi^2))(-Sqrt[sp*(sp - tr)] + Sqrt[sm*(sm - tr)] + (1./2.)*tr*Log[(2.*sp + 2.*Sqrt[sp*(sp-tr)] - tr)/(2.*sm + 2.*Sqrt[sm*(sm - tr)] - tr)]);
            matrix[[i,i]]       =   inter; (*diagonal elements*)
        ];
    ];
    (*###############*)

    Do[
        qj                      =   q[[j]];
        xqj                     =   xq[[j]];

        (* Model dependent*)
        (*###############*)
        ctu                     =   ( xqi*xqj+ xi^2-tr/2.)/(qi*qj);
        If[ctu>-1,
            If[ctu>1,
                sp              =   2. xi^2+2.(xqi xqj -qi qj ),
                sp              =   tr
            ];
            sm                  =   2. xi^2+2.(xqi xqj +qi qj);
            If[sp-tr>=0&&sm-tr>=0&&sm>sp,
                inter           =   (const/(xqi xqj qi qj))(-Sqrt[sp*(sp - tr)] + Sqrt[sm*(sm - tr)] + (1./2.)*tr*Log[(sp + Sqrt[sp*(sp-tr)] - tr/2.)/(sm + Sqrt[sm*(sm - tr)] - tr /2.)]);
                matrix[[i,j]]   =   inter;(*off-diagonal elements*)
                matrix[[j,i]]   =   inter;
            ];
        ];
        (*###############*)
    ,{j,i+1,dim}
    ];
,{i,1,dim}
];

pref*matrix

], RuntimeOptions->"Speed", CompilationTarget->"C"];




GetScatMatrix=Compile[{{pref,_Real,0},{dim,_Integer,0},{c2,_Real,0},{xi,_Real,0},{tq,_Real,1},{param,_Real,1}},Module[{i,j,matrix,qi,qj,xqi,xqj,sq1,sq2,sumq,diffq,a,b,c,d,delta,q,xq,inter,M2,r,const},
(*
INPUT:

    pref                        =   [gDM/(2 Pi^2 tilde H x)]*(m/Sqrt[x])^3 dqA for coordinates A,
    pref                        =   [gDM/(2 Pi^2 tilde H x)]*(m/x)^3 g_s dqB for coordinates B.
    dim                         -   Dimesion of tq
    tq                          -   momentum table
    p/T                         =   Sqrt[c2] tq, where c2 and tq depend on momentum coordinates A and B.
    xi                          =   mDM/T
    param                       =   modelparameters

OUTPUT:

pref*matrix                     - scattering collision term as a matrix


COMMENT:

matrix = \hat{C}_el * mDM^2 2 \Pi^2 /(c2 T^2 \Delta p)
\partial_x f = pref matrix.f + annihilation + ...
Used in fBE if FullCell  = True;
This routine is specialized to the threshold example.
Please modify "model dependent parts" for adopting this routine for other <W> .
*)


matrix                          =   ConstantArray[0.,{dim,dim}];
q                               =   Sqrt[c2] tq;
xq                              =   Sqrt[xi^2+c2*tq^2];

(* Model dependent*)
(*###############*)
M2                              =   param[[2]];
r                               =   param[[3]];
const                           =   M2*xi^2*(Sqrt[r xi/(2.*3.1415)]/64.)Exp[-r xi];
sq1                             =   Sqrt[r*xi/2.];
sq2                             =   1./Sqrt[8. r xi];
(*###############*)

Do[
    qi                          =   q[[i]];
    xqi                         =   xq[[i]];
    Do[
        qj                      =   q[[j]];
        xqj                     =   xq[[j]];

        (* Model dependent*)
        (*###############*)
        sumq                    =   qi+qj;
        diffq                   =   Abs[qi-qj];
        delta                   =   Abs[xqi-xqj];
        a                       =   sq1*delta/diffq;
        b                       =   sq2*diffq;
        c                       =   sq1*delta/sumq;
        d                       =   sq2*sumq;
        inter                   =   const*(1./(xqi*xqj*qi*qj))(Exp[delta/2.](Erfc[a+b]-Erfc[c+d])+Exp[-delta/2.](Erfc[c-d]-Erfc[a-b]));(* This implements mDM^2*<W>_{ij} *)
        (*###############*)
        matrix[[i,j]]           =   tq[[j]]^2 inter;
        matrix[[j,i]]           =   tq[[i]]^2 inter;
    ,{j,i+1,dim}
    ];
,{i,1,dim}
];

Do[
    inter                       =   0.;
    Do[
        inter                   -=  matrix[[i,j]]Exp[(xq[[i]]-xq[[j]])/2.];
    ,{j,1,dim}
    ];
    matrix[[i,i]]               =   inter;
,{i,1,dim}
];

Do[
    Do[
        matrix[[i,j]]           =   matrix[[i,j]] Exp[-((xq[[i]]-xq[[j]]))/2.];
        matrix[[j,i]]           =   matrix[[j,i]]Exp[((xq[[i]]-xq[[j]]))/2.];
    ,{j,i+1,dim}
    ];
,{i,1,dim}
];

pref*matrix

], RuntimeOptions->"Speed", CompilationTarget->"C"];

GetSecondMomentScat=Compile[{{xi,_Real,0},{yeq,_Real,0},{y,_Real,0},{ExpDbk2,_Real,0},{derbk2Dbk2,_Real,0},{param,_Real,1}},Module[{dim,i,j,matrix,yeqdy,xDM,qi,qj,xqi,xqj,sq1,sq2,sumq,diffq,a,b,c,d,delta,q,xq,inter,qmin,qmax,dq,fdm,avscatt,deravscatt,secmom,M2,r,const},
(*
INPUT:

    xi                          =   mDM/T
    yeq                         =   mDM T/s^(2/3)
    y                           =   mDM T_chi/s^(2/3)
    ExpDbk2                     =   Exp[-xDM]/BesselK2[xDM]
    derbk2Dbk2                  =   BesselK2'[xDM]/BesselK2[xDM] -- with prime = d/dxDM
    param                       =   modelparameters

OUTPUT:

{avscatt,deravscatt} -- second momentum moment average of scattering collision term and its derivative with respect to y

COMMENT:

avscatt= < C_el>_2 * T_chi/(m_chi^2)
deravscatt= d avscatt /dy
Used in cBE if FullCell  = True;
This routine is specialized to the threshold example.
Please modify "model dependent parts" for adopting this routine for other <W> .
*)

(* construct moving q grid *)
dim                             =   250;
yeqdy                           =   yeq/y;
qmin                            =   10^(-4)Sqrt[2.xi/yeqdy];
qmax                            =   8.*Sqrt[2. xi/yeqdy];
dq                              =   (qmax-qmin)/(dim-1.);
q                               =   Table[qmin+(k-1.) dq,{k,dim}];
xDM=xi*yeqdy;
matrix                          =   ConstantArray[0.,{dim,dim}];
xq                              =   Sqrt[xi^2+q^2];
fdm                             =   Exp[-(yeqdy*xq-xDM)]*ExpDbk2;


(* Model dependent*)
(*###############*)
M2                              =   param[[2]];
r                               =   param[[3]];
const                           =   M2*xi^2*(Sqrt[r xi/(2.*3.1415)]/64.)Exp[-r xi];
sq1                             =   Sqrt[r*xi/2.];
sq2                             =   1./Sqrt[8. r xi];
(*###############*)

Do[
    qi                          =   q[[i]];
    xqi                         =   xq[[i]];
    Do[
        qj                      =   q[[j]];
        xqj                     =   xq[[j]];

        (* Model dependent*)
        (*###############*)
        sumq                    =   qi+qj;
        diffq                   =   Abs[qi-qj];
        delta                   =   Abs[xqi-xqj];
        a                       =   sq1*delta/diffq;
        b                       =   sq2*diffq;
        c                       =   sq1*delta/sumq;
        d                       =   sq2*sumq;
        inter                   =   const*(1./(xqi*xqj*qi*qj))(Exp[delta/2.](Erfc[a+b]-Erfc[c+d])+Exp[-delta/2.](Erfc[c-d]-Erfc[a-b])); (* This implements mDM^2*<W>_{ij} *)
        (*###############*)
        matrix[[i,j]]           =   q[[j]]^2 inter;
        matrix[[j,i]]           =   q[[i]]^2 inter;
    ,{j,i+1,dim}
    ];
,{i,1,dim}
];

Do[
    inter                       =0.;
    Do[
        inter                   -=  matrix[[i,j]]Exp[(xq[[i]]-xq[[j]])/2.];
    ,{j,1,dim}
    ];
    matrix[[i,i]]               =   inter;
,{i,1,dim}
];

Do[
    Do[
        matrix[[i,j]]           =   matrix[[i,j]] Exp[-((xq[[i]]-xq[[j]]))/2.];
        matrix[[j,i]]           =   matrix[[j,i]]Exp[((xq[[i]]-xq[[j]]))/2.];
    ,{j,i+1,dim}
    ];
,{i,1,dim}];

(*matrix= \hat{C}_el * mDM^2 2 \Pi^2 /( T^2 \Delta p) *)

c                               =   (xDM xi^(-7) dq^2/(6.*3.1415^2));
secmom                          =   q^4/xq;
avscatt                         =   c*((secmom).(matrix.fdm));
deravscatt                      =   ((xDM derbk2Dbk2) -1.)*(avscatt/y)+c*(yeqdy/y)*((secmom).(matrix.(xq*fdm)));

{avscatt,deravscatt}

], RuntimeOptions->"Speed", CompilationTarget->"C"];
