(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Various auxiliary settings and options.
Definitions for some of the usuful functions.

Version: 23-01-2021
Contact: Andrzej Hryczuk (NCBJ)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)

(* for using NLimit *)
(*Needs["NumericalCalculus`"];*)

(* for using packed arrays: *)
Needs["Developer`"];
(* the following allows to use GOTO and LABEL in compiled functions. *)
<< CompiledFunctionTools`;
Unprotect[CompiledFunctionTools`Private`getInstruction];
CompiledFunctionTools`Private`getInstruction[line_, {0, _}]:=
CompiledFunctionTools`Private`getInstruction[line, {3, 1}];

(*---------------------------- auxiliary routines ----------------------------*)

(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Routine for loading the Model.
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)
GetModel[modelname_,parametersname_,settingsname_]:=Module[{},

  Get["./models/"<>modelname<>"/"<>parametersname<>".wl"];
  Get["./models/"<>modelname<>"/"<>settingsname<>".wl"];
  Get["./models/"<>modelname<>"/"<>modelname<>".wl"];

  (* identify the names with global variables *)
  model      =  modelname;
  parameters =  parametersname;
  settings   =  settingsname;

  (* import the degrees of freedom *)
  dofimport[dof];

  (* print out values of input model parameters *)
  Print[cardinfo];
];


(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Routine for running the main.wls script
By: Michael Gustafsson (Added on: 16-02-2021)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)
RunMain[modelname_,parametersname_,settingsname_]:=Module[{},

   model=modelname;
   parameters=parametersname;
   settings=settingsname;
   Get["main.wls"];
 ];


(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Routine for running the main.wls script
By: Michael Gustafsson (Added on: 16-02-2021)
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)
AddModel[modelname_]:=Module[{},

  dir = Directory[];
  SetDirectory["./models"];
  newmodelname=modelname;
  Get["AddModel.wls"];
  SetDirectory[dir];
 ];


(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Routine for making output plots
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)
MakePlots :=
  Module[{tabs,style,legend,Ytabs,Ystyle,Ylegend,ytabs,ystyle,ylegend},

  (*global options for plots*)
   SetOptions[{ListLogLogPlot, ListLogPlot, ListPlot, LogLogPlot},
    ImageSize -> 400, Axes -> False, Frame -> True];

   (****Y evolution plot****)
   tabs = {tYnBE, tYcBE, tYfBE};
   style = {Gray, ColorData[97, 1], Orange};
   legend = {Style["nBE", Gray], Style["cBE", ColorData[97, 1]],Style["fBE", Orange]};
   Ytabs = {};
   Ystyle = {};
   Ylegend = {};
   Do[
    If[Length[tabs[[i]]] != 0,
      AppendTo[Ytabs, tabs[[i]]];
      AppendTo[Ystyle, style[[i]]];
      AppendTo[Ylegend, legend[[i]]];
      ];
    , {i, Length[tabs]}
    ];

    plY = ListLogLogPlot[Ytabs, FrameLabel -> {"x=\!\(\*SubscriptBox[\(m\), \(DM\)]\)/T", "\!\(\*SubscriptBox[\(Y\), \(DM\)]\)"},
      FrameStyle -> Directive[Bold, 12], PlotStyle -> Ystyle,
      PlotLegends -> {Placed[LineLegend[Ystyle, Ylegend,
        LegendMarkers -> Automatic,
        LegendMarkerSize -> 12],
        {Right, Top}]}
      ];

   (* print *)
   Print[plY];

   (****y evolution plot****)
   tabs = {tycBE, tyfBE};
   style = {{ColorData[97, 1]}, Orange};
   legend = {Style["cBE", ColorData[97, 1]], Style["fBE", Orange]};
   ytabs = {};
   ystyle = {};
   ylegend = {};
   Do[
    If[Length[tabs[[i]]] != 0,
      AppendTo[ytabs, tabs[[i]]];
      AppendTo[ystyle, style[[i]]];
      AppendTo[ylegend, legend[[i]]];
      ];
    , {i, 1, Length[tabs]}
    ];

   If[Length[ytabs]!=0,
    ply = Show[
      ListLogLogPlot[ytabs, FrameStyle -> Directive[Bold, 12],
       PlotStyle -> ystyle,
       FrameLabel -> {"x=\!\(\*SubscriptBox[\(m\), \(DM\)]\)/T","\!\(\*SubscriptBox[\(y\), \(DM\)]\)"},
       PlotLegends -> {Placed[
          LineLegend[Join[{Gray}, ystyle],
           Join[{Style["\!\(\*SubscriptBox[\(y\), \(eq\)]\)", Gray]},ylegend],
           LegendMarkers -> Automatic,
           LegendMarkerSize -> 12],
           {Right, Bottom}]}
           ],
      LogLogPlot[yeq[mDM, x], {x, xmin, xmax}, PlotStyle -> Gray]
      ];
      (* print *)
      Print[ply];
    ];

   (****f_DM plot at 3 different times:start,mid,end****)
  If[Length[tYfBE]!=0,
   fDMxi = tfDM[[1, 2]]; xi = Round[tfDM[[1, 1]]];
   fDMxm = tfDM[[2, 2]]; xm = Round[tfDM[[2, 1]]];
   fDMxf = tfDM[[-1, 2]]; xf = Round[tfDM[[-1, 1]]];
   If[IntegerQ[Log10[xf]], xf = "1e" <> ToString[Log10[xf]]];
   plf = ListPlot[{fDMxi, fDMxm, fDMxf}, PlotRange -> All,
     FrameLabel -> {"q", "\!\(\*SubscriptBox[\(f\), \(DM\)]\)"},
     FrameStyle -> Directive[Bold, 12],
     PlotStyle -> {ColorData[97, 1], ColorData[97, 3], ColorData[97, 4]},
     PlotLegends -> {
       Placed[
        LineLegend[{ColorData[97, 1], {ColorData[97, 3]}, ColorData[97, 4]},
        {Style["x=" <> ToString[xi], ColorData[97, 1]],
          Style["x=" <> ToString[xm], ColorData[97, 3]],
          Style["x=" <> ToString[xf], ColorData[97, 4]]},
         LegendMarkers -> Automatic ,
         LegendMarkerSize -> 12],
        {Right,Top}]}];

   (* print *)
   Print[plf];
   ]
];


(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Routine for saving the output.

If save all information from the DRAKE` context use
[Warning: the resulting file may be large and its saving can slow down the code]

  save[outfile,"all"];

if only minimal, i.e. main results without parameters and
full intermediate results

  save[outfile];
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)
save[outfile_,type___]:= Module[{},

 If[!FileExistsQ[outfile],
  Save[outfile,
    {runtime,Oh2nBE,Oh2cBE,Oh2fBE,tYcBE,tycBE,tyfBE,tYnBE,tYfBE,tfDM,
      tsv,svx,tsv2,sv2x,tsvtheta,tgamma,gam}];
 ,
  Print["Result already exists from previous run! Saving as file with extension .tmp"];
  Save[outfile<>".tmp",
    {runtime,Oh2nBE,Oh2cBE,Oh2fBE,tYcBE,tycBE,tyfBE,tYnBE,tYfBE,tfDM,
      tsv,svx,tsv2,sv2x,tsvtheta,tgamma,gam}];
 ];

 (* if called without "min" option add on top of that all the other symbols in the context *)
 If[type!="min", Save[outfile,"DRAKE`"];];

];

(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Routine for determining the x value for which gam(x)/H(x)==gamcap
set in the settings file. Then this x value is used as a starting point for
fBE calculation, i.e. or x<xcap the code assumes kinetic equilibrium and
solves the nBE
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)
Getxcap := Module[{xcap=0.,f,df,xcaptab,i=1,xcapint},

  (* looking for x where f[x]==1 *)
  f[x_]  := gam[x]/(H[mDM/x])/gamcap;
  df[x_] := f[x + 0.01] - f[x];

  (* tabulate and interpolate to improve FindRoot *)
  xcaptab = Table[{x, f[x]}, {x, 10.^Range[Log10[xmin], Log10[xmax], 0.01]}];
  xcapint = Interpolation[xcaptab, InterpolationOrder -> 1];

  (* check if there is a possibility of reaching 1 *)
  If[f[xmin] < 1 && df[xmin] < 0,
    (* if not, then return -1 *)
    xcap = -1
  ,
    (* find index where f[x] ~ 1 to bracket initial condition for FindRoot *)
    While[xcaptab[[i, 2]] > 1, i++];
    xcap = x /. FindRoot[Log[xcapint[x]] == 0, {x,(xcaptab[[i-1,1]]+xcaptab[[i,1]])/2, xcaptab[[i-1,1]], xcaptab[[i,1]]}][[1]]
  ]
 ];

(* %%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%
Other usuful functions/routines/quntities/options
%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%%% *)

(*------------------------------ Warning options -----------------------------*)
SwitchOffWarnings:=Module[{},
  Off[FrontEndObject::notavail]; (* no front end *)
  Off[General::munfl];  (* loss of precision *)
  Off[Infinity::indet]; (* Indeterminate *)
  (* convergence and accuracy warnings for integrations *)
  Off[NIntegrate::eincr,NIntegrate::ncvb,NIntegrate::izero,NIntegrate::slwcon];
  ParallelEvaluate[Off[NIntegrate::eincr,NIntegrate::ncvb,NIntegrate::izero,NIntegrate::slwcon]];
];

SwitchOnWarnings:=Module[{},
  On[FrontEndObject::notavail]; (* no front end *)
  On[General::munfl];  (* loss of precision *)
  On[Infinity::indet]; (* Indeterminate *)
  (* convergence and accuracy warnings for integrations *)
  On[NIntegrate::eincr,NIntegrate::ncvb,NIntegrate::izero,NIntegrate::slwcon];
  ParallelEvaluate[On[NIntegrate::eincr,NIntegrate::ncvb,NIntegrate::izero,NIntegrate::slwcon]];
];

 (* --------------------------------- Graphics --------------------------------*)
 (* Messages in a C form better suited for terminal output *)
 If[$FrontEnd === Null,$MessagePrePrint=CForm];
 (* print style in color only when in Notebook environment *)
 Stylef[x_,y_] := If[$FrontEnd === Null,x,Style[x,y]];



 (*-------------------------------- Constants ---------------------------------*)
 MPl         = 1.2209 10^(19)/(Sqrt[8Pi]);
 Kelvin      = 8.617343*10^(-14);                         (* Kelvin in GeV *)
 Meter       = 1/(197.326961*10^(-18)) ;                  (* Meter in 1/GeV *)
 Kilogramm   = 0.510998910*10^(-3)/(9.10938215*10^(-31)); (* Kilogramm in GeV *)
 Ttoday      = 2.725 Kelvin;                              (* T today in Kelvin *)
 Rhocritical = 1.05368 10^(-5)/(0.01 Meter)^3;            (* critical density/h^2*)
 Entropytoday= iheff[Ttoday]*2 Pi^2/45*Ttoday^3;          (* entropy today *)

 (* replace 0 by the minimal number when may give rise to 1/0 *)
 minnumber = 10.^-100;

 (*---------------------------------- Functions -------------------------------*)
 (* equilibrium solutions: *)
 neq[m_,x_] := m^3 gDM BesselK2[x]/(2 Pi^2 x);  (* n_equilibrium of the DM; (needs gDM to be specified in the model file) *)
 feq[x_,q_] := Exp[-Sqrt[x^2 + q^2]]; (* the equlibrium f(x,q) *)
 Yeq[m_,x_] := neq[m,x]/s[m/x]; (* Y_equilibrium of the DM *)
 yeq[m_,x_] := m^2/(x) s[m/x]^(-2/3); (* y_equilibrium *)

 (* very well approximated expanded Bessel functions: *)
 xBexp = 100;
 BesselK1[x_ /; x < xBexp]  := BesselK[1, x];
 BesselK1[x_ /; x >= xBexp] := Exp[-x]*Sqrt[Pi/(2*x)]*(1 + 72765/(262144*x^5) -
                     4725/(32768*x^4) + 105/(1024*x^3) - 15/(128*x^2) + 3/(8*x));
 BesselK2[x_ /; x < xBexp]  := BesselK[2, x];
 BesselK2[x_ /; x >= xBexp] := Exp[-x]*Sqrt[Pi/(2*x)]*(1 - 135135/(262144*x^5) +
                   10395/(32768*x^4) - 315/(1024*x^3) +105/(128*x^2) + 15/(8*x));
 (* Exp[-x]/K_2: *)
 ExpDbk2[x_ /; x < xBexp]  := Exp[-x]/BesselK[2, x];
 ExpDbk2[x_ /; x >= xBexp] := 1/(Sqrt[Pi/(2*x)]*(1 - 135135/(262144*x^5) +
                  10395/(32768*x^4) - 315/(1024*x^3) +105/(128*x^2) + 15/(8*x)));
 (* K'_2/K_2 : *)
 Derbk2Dbk2[x_ /; x < xBexp] := 1/2((-BesselK[1, x]-BesselK[3, x])/BesselK[2, x]);
 Derbk2Dbk2[x_ /; x >= xBexp] :=-1. - x^(-1)/2. - (15.*x^-2)/8. + (15.*x^-3)/8.
                       - (135.*x^-4)/128. - (45.*x^-5)/32. + (7425.*x^-6)/1024.;
